"""Run the app."""
import sys

from signal import signal, SIGINT

from app.flask_app import create_app


def exit_handle(sig, frame):
    """Exit when CTRL+C is pressed."""
    sys.exit(0)


def main():
    """Run the app."""
    app = create_app()

    port_range = (50000, 51000)
    for port in range(*port_range):

        signal(SIGINT, exit_handle)

        try:
            app.run(debug=True, port=port)
        except Exception:
            if port == max(port_range) - 1:
                raise RuntimeError(f"All ports in range {port_range} in use")
            # need some way to make sure it doesn't re-run
            continue


if __name__ == "__main__":
    main()



"""Function to run depht."""
import pathlib
import shutil

from datetime import datetime
from tempfile import mkdtemp

from Bio import SeqIO

from depht.functions.annotation import (annotate_record,
                                        cleanup_flatfile_records)
from depht.functions.sniff_format import sniff_format
from depht.functions.mmseqs import assemble_bacterial_mask
from depht.functions.multiprocess import CPUS
from depht.functions.find_homologs import find_homologs
from depht.functions.prophage_prediction import predict_prophage_coords
from depht.data import GLOBAL_VARIABLES, PARAMETERS
from depht.__main__ import (load_contigs, load_initial_prophages,
                            detect_att_sites, write_prophage_output)


# ------------ GLOBAL VARIABLES ---------------
DEPHT_DIR = pathlib.Path().home().joinpath(
                                GLOBAL_VARIABLES["model_storage"]["home_dir"])
if not DEPHT_DIR.is_dir():
    DEPHT_DIR.mkdir()

MODEL_DIR = DEPHT_DIR.joinpath(GLOBAL_VARIABLES["model_storage"]["model_dir"])
if not MODEL_DIR.is_dir():
    MODEL_DIR.mkdir()

# Where temporary file I/O should happen
TMP_DIR = DEPHT_DIR.joinpath(GLOBAL_VARIABLES["model_storage"]["tmp_dir"])

# Model can't scan contigs with fewer CDS than window size
MIN_CDS_FEATURES = PARAMETERS["prophage_prediction"]["window"]
# For naming any identified prophages
PROPHAGE_PREFIX = GLOBAL_VARIABLES["phage_sequences"]["prophage_prefix"]
PROPHAGE_DELIMITER = GLOBAL_VARIABLES["phage_sequences"]["prophage_delimiter"]

# For attachment site detection
EXTEND_BY = PARAMETERS["att_detection"]["extention_length"]
ATT_SENSITIVITY = PARAMETERS["att_detection"]["att_sensitivity"]

# For deciding whether to cull predicted "prophages"
MIN_LENGTH = PARAMETERS["annotation"]["min_length"]
MIN_PRODUCTS_NORMAL = PARAMETERS["phage_homology_search"]["min_products"]
MIN_PRODUCTS_STRICT = PARAMETERS["phage_homology_search"][
                                 "min_products_strict"]


def run_depht(infiles, outdir, product_threshold, model,
              cpus=CPUS, draw=True, dump=False, verbose=False,
              runmode="normal", att_sens=ATT_SENSITIVITY, tmpdir=TMP_DIR,
              min_length=MIN_LENGTH):
    """Run DEPhT."""
    # Make sure output dir is a valid path
    if not outdir.is_dir():
        print(f"'{str(outdir)}' does not exist - creating it...")
        outdir.mkdir(parents=True)

    # Get the temporary directory, refresh it if it exists
    if not tmpdir.is_dir():
        tmpdir.mkdir(parents=True)

    # Set up model directories
    model_dir = MODEL_DIR.joinpath(model)

    shell_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["shell_db"]["dir_name"])
    reference_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["reference_db"]["dir_name"])
    phage_homologs_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["phage_homologs"]["dir_name"])

    bact_ref_fasta = shell_db_dir.joinpath(
                        GLOBAL_VARIABLES["shell_db"]["fasta_name"])
    bact_ref_values = shell_db_dir.joinpath(
                        GLOBAL_VARIABLES["shell_db"]["hex_value_name"])
    essential_db = phage_homologs_dir.joinpath(
                        GLOBAL_VARIABLES["phage_homologs"]["essential_name"])
    extended_db = phage_homologs_dir.joinpath(
                        GLOBAL_VARIABLES["phage_homologs"]["extended_name"])
    blast_db = reference_db_dir.joinpath(
                        GLOBAL_VARIABLES["reference_db"]["name"])
    classifier = model_dir.joinpath(
                        GLOBAL_VARIABLES["classifier"]["name"])

    # Mark program start time
    mark = datetime.now()

    # OK, let's go!
    for infile in infiles:
        # Make sure filepath exists - skip if it doesn't
        if not infile.is_file():
            print(f"'{str(infile)}' does not exist - skipping it...")
            continue

        # Skip .DS_Store files
        if infile.name == ".DS_Store":
            print("skipping .DS_Store file...")
            continue

        # Set up a temporary directory for this genome
        genome_tmp_dir = pathlib.Path(mkdtemp(dir=tmpdir))
        if not genome_tmp_dir.is_dir():
            genome_tmp_dir.mkdir()

        # Automatically check the file format
        fmt = sniff_format(infile)

        # Parse all contigs of annotation-worthy length
        if verbose:
            print(f"\nparsing '{str(infile)}' as {fmt}...")

        # Parse the file and retain contigs above cutoff length
        records = [x for x in SeqIO.parse(infile, fmt) if len(x) >= MIN_LENGTH]

        if not records:
            if verbose:
                print(f"no {fmt}-formatted records of at least {MIN_LENGTH}bp "
                      f"found in '{str(infile)}' - skipping it...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        # Annotate contigs if format is "fasta"
        if fmt == "fasta":
            if verbose:
                print("annotating t(m)RNA and CDS genes de novo...")

            annotate_dir = genome_tmp_dir.joinpath("annotate")
            if not annotate_dir.is_dir():
                annotate_dir.mkdir()

            for record in records:
                annotate_record(record, annotate_dir)

        else:
            if verbose:
                print("using flat file annotation...")

            cleanup_flatfile_records(records)

        # Filter contigs that don't have enough CDS features
        records = [record for record in records if (len(
            [x for x in record.features
                if x.type == "CDS"]) >= MIN_CDS_FEATURES)]

        if not records:
            print(f"no contigs with enough CDS features to analyze in "
                  f"'{str(infile)}' - skipping it...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        contigs = load_contigs(records)

        if verbose:
            print("masking conserved bacterial features...")

        mmseqs_dir = genome_tmp_dir.joinpath("mmseqs")
        if not mmseqs_dir.is_dir():
            mmseqs_dir.mkdir()

        # Detect conserved bacterial genes for each contig
        bacterial_masks = assemble_bacterial_mask(
            contigs, bact_ref_fasta, bact_ref_values, mmseqs_dir)

        if verbose:
            print("looking for high-probability prophage regions...")

        # Predict prophage coordinates for each contig
        prophage_predictions = list()
        for i, contig in enumerate(contigs):
            prediction = predict_prophage_coords(
                contig, classifier, EXTEND_BY, mask=bacterial_masks[i])

            filtered_prediction = []
            for pred in prediction:
                if len(range(*pred)) < min_length:
                    continue

                filtered_prediction.append(pred)

            prophage_predictions.append(filtered_prediction)

        if all([not any(x) for x in prophage_predictions]) and not dump:
            print(f"no complete prophages found in '{str(infile)}'...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        product_threshold = 0
        if runmode in ("normal", "sensitive"):
            hhsearch_dir = genome_tmp_dir.joinpath("hhsearch")
            if not hhsearch_dir.is_dir():
                hhsearch_dir.mkdir()

            # Search for phage gene remote homologs
            if verbose:
                print("searching for phage gene homologs...")

            find_homologs(contigs, prophage_predictions, essential_db,
                          hhsearch_dir, cpus)
            product_threshold = MIN_PRODUCTS_NORMAL

            if runmode == "sensitive":
                if verbose:
                    print("extending search for phage gene homologs...")
                find_homologs(contigs, prophage_predictions, extended_db,
                              hhsearch_dir, cpus, cache_scores=False)
                product_threshold = MIN_PRODUCTS_STRICT
        else:
            for contig in contigs:
                contig.fill_hhsearch_scores()

        prophages = load_initial_prophages(contigs, prophage_predictions,
                                           product_threshold,
                                           prefix=PROPHAGE_PREFIX,
                                           delimiter=PROPHAGE_DELIMITER)

        if verbose and prophages:
            print("searching for attL/R...")

        # Set up directory where we can do attL/R detection
        att_dir = genome_tmp_dir.joinpath("att_core")
        if not att_dir.is_dir():
            att_dir.mkdir()

        # Detect attachment sites, where possible, for the predicted prophage
        search_space = att_sens * EXTEND_BY
        detect_att_sites(prophages, blast_db, search_space, att_dir)

        prophages = [prophage for prophage in prophages
                     if prophage.length >= min_length]

        if not prophages and not dump:
            print(f"no complete prophages found in '{str(infile)}'...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        if verbose:
            print("generating final reports...")

        genome_outdir = outdir.joinpath(f"{infile.stem}")
        if not genome_outdir.is_dir():
            genome_outdir.mkdir()

        draw_dir = genome_tmp_dir.joinpath("draw_diagram")
        if not draw_dir.is_dir():
            draw_dir.mkdir()
        write_prophage_output(genome_outdir, contigs, prophages, draw_dir,
                              draw)

        if dump:
            destination = genome_outdir.joinpath("tmp_data")

            if destination.exists():
                shutil.rmtree(destination)

            shutil.copytree(genome_tmp_dir, destination)
        shutil.rmtree(genome_tmp_dir)  # clean up after ourselves

    print(f"\nTotal runtime: {str(datetime.now() - mark)}")



"""DEPhT."""
import shutil
import webbrowser

from datetime import datetime
from pathlib import Path
from flask import (Blueprint, render_template, current_app, request,
                   send_from_directory)
from run_depht.run_depht import run_depht

bp = Blueprint("depht", __name__)


def check_dir(dirpath):
    """Check if directory exists, otherwise make it."""
    if not dirpath.is_dir():
        dirpath.mkdir(parents=True)


def make_output_dir():
    """Create directory structure for DEPhT output.
    Directory structure is as follows:
    home/Downloads/depht/<date>/<time>
    Output stored within this directory
    """
    # timestamp
    timestamp = str(datetime.now())

    # date
    date = timestamp.split(" ")[0]
    date_list = date.split("-")

    time = timestamp.split(" ")[1]
    time_list = time.split(":")
    time_list[2] = str(round(float(time_list[2])))
    time = ":".join(time_list)

    month = date_list[1]
    month_name = datetime.strptime(month, "%m")
    month_name = month_name.strftime("%b")

    date_dir_name = (date_list[0] + "-" + month_name + "-" + date_list[2])

    # outdir = Path.home()/"Downloads"/("depht_" + timestamp)

    depht_outdir = Path.home()/"Downloads"/"depht"

    outdir = depht_outdir/date_dir_name/time

    return outdir


def no_model_redirect(window_size, cpu_cores, model_url):
    """If no local models are saved, redirect.

    Opens a new tab with the models and also redirects to training page.
    """
    webbrowser.open_new_tab(model_url)

    return render_template("train/train.html", no_models=True,
                           window=window_size, cpu_cores=cpu_cores)


@bp.route("/")
def depht():
    """Render depht."""
    length = current_app.config["MIN_LENGTH"]
    cpu_cores = current_app.config["CPUS"]
    window_size = current_app.config["WINDOW"]
    att = current_app.config["ATT_SENSITIVITY"]
    model_dir = current_app.config["MODEL_DIR"]
    model_url = current_app.config["MODEL_URL"]

    # local_models = current_app.config["LOCAL_MODELS"]

    local_models = [model.name for model in model_dir.iterdir()
                    if model.is_dir()]

    # if there are no local models stored
    # open a new tab with the osf path
    # simultaneously open to train models page
    if len(local_models) == 0:
        return no_model_redirect(window_size, cpu_cores, model_url)

    return render_template("depht/depht.html",
                           length=length,
                           cpu_cores=cpu_cores,
                           att=att,
                           local_models=local_models)


@bp.route("/results", methods=["POST"])
def depht_results():
    """Display the results."""
    # infiles = [Path(request.args.get("input_file"))]
    # infile = infiles[0]
    # timestamp
    outdir = make_output_dir()

    infiles = request.files.getlist("input_file")

    input_dir = outdir/"input_files"
    check_dir(input_dir)

    # make sure there are no already existing input_files directories
    for dir in input_dir.iterdir():
        if dir.is_dir() and "input_file" in dir.name:
            shutil.rmtree(dir)

    print(list(request.files.items()))

    for file in infiles:
        file.save(Path(input_dir/file.filename))

    infiles = list(input_dir.iterdir())

    # outdir = Path(str(infile.stem) + "_output")
    # outdir = Path(request.args.get("output_folder"))

    model = request.form["model"]
    runmode = request.form["mode"]
    att_sens = int(request.form["att"])
    products = int(request.form["products"])
    length = int(request.form["length"])
    cpus = int(request.form["cores"])

    try:
        dump_data = request.form["dump"]
    except Exception:
        dump_data = False

    # always draw the html map
    draw = True

    args = {"Model": model,
            "Runmode": runmode,
            "Att Sensitivity parameter": att_sens,
            "Min. phage homologs": products,
            "Min. length": length,
            "CPU cores": cpus}

    run_depht(infiles, outdir, products, model, cpus=cpus, draw=draw,
              runmode=runmode, dump=dump_data, att_sens=att_sens,
              min_length=length)

    # shutil.rmtree(input_dir)

    # filepaths = []

    # output
    # path to the html file if the draw option was selected
    # html file will be embedded in the results page
    # other files will be linked to their filepath in the user's outdir

    # output written to a directory named for the input file in outdir
    # if directory doesn't exist, no prophages were found

    prophage_table = {}

    for infile in infiles:
        # infile = infiles[0]
        results_dir = outdir/infile.stem

        html_file = None
        proph_count = 0
        # no prophage found if the results directory does not exist
        # or if there is no html file in the results directory
        if not results_dir.is_dir():
            prophage_table[infile.stem] = proph_count
            continue
        else:
            suffix_list = [file.suffix for file in results_dir.iterdir()]
            if ".html" not in suffix_list:
                prophage_table[infile.stem] = proph_count
                continue

        proph_count = 0
        # html_files = []

        for item in results_dir.iterdir():
            # found a prophage
            if item.is_dir() and item.name != "tmp_data":
                proph_count += 1

            '''
            # will display the html if there is one input file only
            if len(infiles) == 1:
                if item.suffix == ".html":
                    html_file = str(item)
            '''
            if item.suffix == ".html":
                html_file = str(item)
                # html_files.append(html_file)

        prophage_table[infile.stem] = proph_count

        if len(infiles) != 1:
            html_file = None

        webbrowser.open_new_tab(str(outdir))

        if html_file:
            outdir = str(results_dir)
        else:
            outdir = str(outdir)

    return render_template("depht/results.html", args=args,
                           table=prophage_table, outdir=outdir,
                           html_file=html_file)


@bp.route("/<path:filename>", methods=["GET"])
def html_display(filename):
    """Display prophage map."""
    filename = Path(filename).name
    dir = request.args.get("dir")

    return send_from_directory(dir, filename, as_attachment=False)


@bp.route("/<path:filepath>", methods=["POST"])
def output_link(filepath):
    """Display a link to where the output is stored."""
    dir = filepath.parent
    filename = filepath.stem
    return send_from_directory(dir, filename, as_attachment=True)



"""Flask app for DEPhT."""
from pathlib import Path
from flask import Flask

from importlib.resources import files

# import app
from app.views import depht, train, models, documentation
from app import config
from app import templates, static
# from app import static, templates

PKG = "depht_gui"

# templates_module = import_module("templates")
# static_module = import_module("static")


def create_app(name=PKG, **kwargs):
    """
    Create the Flask app.

    :param name: name of the app
    :type name: str
    """
    app = Flask(name, template_folder=files(templates),
                static_folder=files(static))

    # app = Flask(name)
    app.config.from_object(config)

    path = Path(app.instance_path)

    if not path.is_dir():
        path.mkdir(parents=True)

    app.register_blueprint(depht.bp)
    app.register_blueprint(train.bp)
    app.register_blueprint(models.bp)
    app.register_blueprint(documentation.bp)

    return app



"""Config file."""
from depht.__main__ import (MIN_CDS_FEATURES, ATT_SENSITIVITY,
                            MIN_LENGTH, MIN_PRODUCTS_NORMAL,
                            MIN_PRODUCTS_STRICT, CPUS, MODEL_DIR, LOCAL_MODELS)
from depht_train.pipelines.train_model import WINDOW

MIN_CDS_FEATURES = MIN_CDS_FEATURES
ATT_SENSITIVITY = ATT_SENSITIVITY

MIN_LENGTH = MIN_LENGTH
MIN_PRODUCTS_NORMAL = MIN_PRODUCTS_NORMAL
MIN_PRODUCTS_STRICT = MIN_PRODUCTS_STRICT

WINDOW = WINDOW
CPUS = CPUS

MODEL_DIR = MODEL_DIR
if not MODEL_DIR.is_dir():
    MODEL_DIR.mkdir(parents=True)

MODEL_URL = "https://osf.io/zt4n3/"

# check that .zip files etc are not added
LOCAL_MODELS = [
    model for model in LOCAL_MODELS if not len(model.split(".")) > 1]



"""DEPhT Train."""
import shutil
import webbrowser

from pathlib import Path
from flask import (Blueprint, render_template, current_app, request,
                   send_from_directory)

from depht_train.pipelines.train_model import train_model
from app.views.depht import check_dir

bp = Blueprint("train", __name__)

CWD = Path.cwd()


@bp.route("/depht-train", methods=["GET"])
def depht_train():
    """Render depht_train page."""
    window_size = current_app.config["WINDOW"]
    cpu_cores = current_app.config["CPUS"]

    return render_template("train/train.html",
                           window=window_size,
                           cpu_cores=cpu_cores)


@bp.route("/model", methods=["POST"])
def training_results():
    """Run depht_train."""
    # get arguments to train a new model
    model_dir = current_app.config["MODEL_DIR"]

    model_name = request.form["name"]
    window = int(request.form["window"])
    cores = int(request.form["cpu"])

    # get the files from the upload, put into temporary directories for input
    # temporary directory is created outside src

    phages = request.files.getlist("phage")
    bacteria = request.files.getlist("bact")
    prophage_csv = request.files["csv"]

    input_dir = CWD.parents[1]/"tmp_infiles"
    phage_dir = input_dir/"phage_dir"
    bact_dir = input_dir/"bact_dir"

    check_dir(phage_dir)
    check_dir(bact_dir)

    for phage_file in phages:
        phage_file.save(Path(phage_dir/phage_file.filename))

    for bact_file in bacteria:
        bact_file.save(Path(bact_dir/bact_file.filename))

    try:
        prophage_csv.save(Path(input_dir/prophage_csv.filename))
        prophage_csv = Path(input_dir/prophage_csv.filename)
        csv_name = prophage_csv.name
    except IsADirectoryError:
        # if the user does not specify a csv file
        prophage_csv = None
        csv_name = None

    # train_model

    args = {"Model name": model_name,
            "Prophage CSV": csv_name,
            "Window Size": window,
            "CPU cores": cores}

    # train the model!
    train_model(model_name, phage_dir, bact_dir, window=window,
                prophages=prophage_csv, cpus=cores)

    # clean up
    shutil.rmtree(input_dir)

    model_dir = model_dir/model_name

    pdf_files = []

    for file in model_dir.iterdir():
        if file.suffix == ".pdf":
            pdf_files.append(file)

    webbrowser.open_new(model_dir)

    return render_template("train/results.html", args=args,
                           model_dir=model_dir,
                           display_files=pdf_files)


@bp.route("/<path:filename>", methods=["GET"])
def pdf_display(filename):
    """Display histograms."""
    filename = Path(filename).name
    dir = request.args.get("dir")

    return send_from_directory(dir, filename, as_attachment=False)



"""Models page."""
import shutil
import webbrowser

from flask import (Blueprint, render_template, request, current_app,
                   send_from_directory)
from pathlib import Path

from app.views.depht import no_model_redirect

bp = Blueprint("models", __name__)


@bp.route("/models")
def models():
    """Render Models page."""
    window_size = current_app.config["WINDOW"]
    cpu_cores = current_app.config["CPUS"]
    model_dir = current_app.config["MODEL_DIR"]
    model_url = current_app.config["MODEL_URL"]
    local_models = [model.name for model in model_dir.iterdir()
                    if model.is_dir()]

    if len(local_models) == 0:
        return no_model_redirect(window_size, cpu_cores, model_url)

    return render_template("models/models.html", local_models=local_models)


@bp.route("/models", methods=["POST"])
def models_action():
    """Carry out the models action."""
    action = request.form["model_action"]
    model_dir = current_app.config["MODEL_DIR"]
    window_size = current_app.config["WINDOW"]
    cpu_cores = current_app.config["CPUS"]
    model_url = current_app.config["MODEL_URL"]

    local_models = [model.name for model in model_dir.iterdir()
                    if model.is_dir()]

    if len(local_models) == 0:
        return no_model_redirect(window_size, cpu_cores, model_url)

    model = request.form["model"]

    # selected model
    model_dir = model_dir/model

    msg = None
    pdf_files = []
    if action == "delete":
        shutil.rmtree(model_dir)

        msg = f"{model} deleted."
        # pdf_files = None
        model_dir = current_app.config["MODEL_DIR"]
        local_models = [model.stem for model in model_dir.iterdir()
                        if model.is_dir()]
    else:
        for file in model_dir.iterdir():
            if file.suffix == ".pdf":
                pdf_files.append(file)
        if len(pdf_files) == 0:
            msg = f"No information found on model {model}."

    print(f"msg = {msg}")

    return render_template("models/models.html", local_models=local_models,
                           msg=msg, pdf_files=pdf_files)


@bp.route("/<path:filename>", methods=["GET"])
def pdf_display(filename):
    """Display histograms."""
    filename = Path(filename).name
    dir = request.args.get("dir")

    return send_from_directory(dir, filename, as_attachment=False)



"""Documentation page."""
from flask import Blueprint, render_template

bp = Blueprint("documentation", __name__)


@bp.route("/documentation")
def documentation():
    """Render Home page."""
    return render_template("documentation.html")
